/**
 * 11kd -- IEEE 802.11k daemon program.
 *
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

#include "netfunc.h"
#include "protocol.h"

#include <stdio.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <signal.h>
#include <errno.h>
#include <syslog.h>
#include <time.h>
#include <mqueue.h>

#include <getopt.h>

#define DEF_LOCK_FILE "11kd.lock"
#define DEF_DOT11KD_MQNAME "/mq11kd"
#define DEF_MQ_TO 1000
#define DEF_PORT 6666
#define DEF_WT_PATH "/opt/wireless-dev/hostap/wlantest/wlantest_cli"

/* options */
int is_daemon;
short port;
char lockfile[128];
char wt_path[128];
int mq_to;
char mq_name[128];
char ap_mac[18];

mqd_t mq;

void log_message(int type, char *message) {
	syslog(type, "%s", message);
}

void print_usage() {
	fprintf(stderr, "11kd -- IEEE 802.11k daemon.\n");
	fprintf(stderr, "Usage: ./11kd [OPTIONS]\n");
	fprintf(stderr, "-d\t--daemonize\t\tRun as a daemon (default: YES).\n");
	fprintf(stderr, "-m\t--mac-addr\t\tAP MAC address\n");
	fprintf(stderr, "-p\t--port\t\t\tPort to listen for messages from collectord (default: 6666)\n");
	fprintf(stderr, "-w\t--wt-path\t\tPath to the wlantest executable\n");
	fprintf(stderr, "-l\t--lock-file\t\tLock file path (default: ./11kd.lock).\n");
	fprintf(stderr, "-q\t--mq-name\t\tMessage queue name (default: /mq11kd).\n");
	fprintf(stderr, "-t\t--mq-timeout\t\tTime to wait for a message on message queue (default: 1000 (msec)).\n");
	fprintf(stderr, "-h\t--help\t\t\tDisplay this message\n");
}

/**
 * Signal handler
 */
void term_handler(int signal) {
	if (unlink(lockfile)) {
		log_message(LOG_ERR, "Error removing lock file");
	}

	closelog();

	mq_close(mq);
	mq_unlink(mq_name);

	exit(0);
}

/**
 * Run as a daemon.
 */
void daemonize() {
	pid_t pid;
	pid_t sid;
	int n;

	int lock_fd;

	char str[10];
	if (getppid() == 1) {
		return;
	}

	pid = fork();
	if (pid < 0) {
		exit(1); /* failure */
	}

	if (pid > 0) {
		exit(0); /* success */
	}

	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	umask(0); /* set newly created file permissions */

	sid = setsid(); /* obtain a new process group */
	if (sid < 0) {
		exit(1);
	}

	lock_fd = open(lockfile, O_RDWR|O_CREAT, 0640);
	if (lock_fd < 0) {
		log_message(LOG_ERR, "Could not create lock file");
		exit(1);
	}

	if (lockf(lock_fd, F_TLOCK, 0) < 0) {
		log_message(LOG_ERR, "Cannot lock pid file. 11kd already running?");
		exit(1);
	}

	sprintf(str, "%d\n", getpid());
	n = write(lock_fd, str, strlen(str));

	signal(SIGCHLD, SIG_IGN); /* ignore child */
	signal(SIGTSTP, SIG_IGN); /* ignore tty signals */
	signal(SIGTTOU, SIG_IGN);
	signal(SIGTTIN, SIG_IGN);
	signal(SIGTERM, term_handler);
	signal(SIGKILL, term_handler);
	signal(SIGINT, term_handler);
	signal(SIGHUP, term_handler);
}

unsigned char **get_sta_list(char *bss, int *n) {
	FILE *fp;
	char cmd[80];
	char line[80];
	unsigned char mac[6];
	unsigned char **retval;
	int num;
	char *ptr;

	memset(cmd, 0, 80);
	memset(line, 0, 80);

	sprintf(cmd, "echo \"list_sta %s\" | %s", bss, wt_path);
	
	fp = popen(cmd, "r");
	if (!fp) {
		*n = 0;
		return NULL;
	}

	/* skip first line of output */
	ptr = fgets(line, 80, fp);

	num = 0;
	retval = NULL;

	while(!feof(fp)) {
		memset(line, 0, 80);
		if(fgets(line, 19, fp) != NULL) {
			if (sscanf(line,
			    "%2hhx:%2hhx:%2hhx:%2hhx:%2hhx:%2hhx", 
			    &mac[0], &mac[1], &mac[2], &mac[3], &mac[4], &mac[5]) == 6) {
				/* parse succeeded */
				if (num == 0) {
					retval = (unsigned char**)malloc(sizeof(unsigned char*));
				}
				else {
					retval = (unsigned char**)realloc(retval, (num+1)*sizeof(unsigned char*));
				}

				retval[num] = (unsigned char*)malloc(6);
				memcpy(retval[num], mac, 6);
				num++;
			}
		}
	}

	pclose(fp);

	*n = num;
	return retval;
}

void send_bcn_req(char *bss, char *sta) {
	FILE *fp;
	char cmd[256];
	char line[80];
	char *ptr;

	memset(cmd, 0, 256);

	sprintf(cmd, "echo \"bcnreq %s %s\" | %s", bss, sta, wt_path);
	fp = popen(cmd, "r");
	if (!fp) {
		return;
	}

	/* dummy fgets */
	ptr = fgets(line, 80, fp);
	pclose(fp);
}

char* get_bss_list() {
	//generate a bss list
	int i;
	struct bss_info_t info[5];
	struct bss_info_t *cur;// = &(info[0]);
		
	for (i = 0; i < 5; i++) {
		cur = &(info[i]);
		if (i == 4) {
			cur->next = NULL;
		}
		else {
			cur->next = &(info[i + 1]);
		}

		generate_random_mac(info[i].bss_mac);
		generate_random_mac(info[i].user_mac);
		info[i].channel = i + 7;
	}

 	return generate_report_batch(&(info[0]));
}

char *handle_collector_request() {
	int n = 0;
	int i = 0;
	int j = 0;
	unsigned int prio;
	unsigned char msgcontent[2048];
	unsigned char *sta_mac;
	unsigned char *bss_mac;
	struct bss_info_t *bss_list;
	struct bss_info_t *cur;
	struct bss_info_t *newnode;
	char *retval;
	struct timespec mq_timeout;

	bss_list = NULL;
	cur = NULL;
	newnode = NULL;


	unsigned char **sta_list = get_sta_list(ap_mac, &n);
	char sta[18];

	/* for each associated station, send a beacon request */
	for (i = 0; i < n; i++) {
		memset(sta, 0, 18);
		sprintf(sta, "%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x", sta_list[i][0], sta_list[i][1], sta_list[i][2], sta_list[i][3], sta_list[i][4], sta_list[i][5]);
		free(sta_list[i]);

		send_bcn_req(ap_mac, sta);		
	}

	if (sta_list) free(sta_list);

	if (mq != (mqd_t)-1) {
		/* now start reading responses from message queue */
		/* we expect to read n messages, one for each client */
		for (i = 0; i < n; i++) {
			memset(msgcontent, 0, 2048);
			mq_timeout.tv_sec = mq_to / 1000;
			mq_timeout.tv_nsec = (mq_to % 1000) * 1000000;

			if (mq_timedreceive(mq, msgcontent, 2048, &prio, &mq_timeout) != -1) {

				sta_mac = msgcontent;
				fprintf(stderr, "Reporter: %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x\n", sta_mac[0], sta_mac[1], sta_mac[2], sta_mac[3], sta_mac[4], sta_mac[5]);
				fprintf(stderr, "Number of BSSs: %u\n", (int)*(msgcontent + 6)); /* TODO: CHECKIT */
				bss_mac = msgcontent + 6 + sizeof(int);
				for (j = 0; j < (int)*(msgcontent + 6); j++) {

					newnode = (struct bss_info_t *)malloc(sizeof(struct bss_info_t));
					memset(newnode, 0, sizeof(struct bss_info_t));				
					memcpy(newnode->user_mac, sta_mac, 6);
					memcpy(newnode->bss_mac, bss_mac, 6);
					newnode->channel = (int)bss_mac[6];
					newnode->next = NULL;

					if (!bss_list) {
						bss_list = newnode;
					}
					else {
						cur->next = newnode;
					}
					cur = newnode;

					fprintf(stderr, "BSS[%d]: %2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x/%d\n", j, bss_mac[0], bss_mac[1], bss_mac[2], bss_mac[3], bss_mac[4], bss_mac[5], (int)bss_mac[6]);
					bss_mac += 7;
				}
			}
		}	
	}

	retval = generate_report_batch(bss_list);

	if (bss_list) {
		cur = bss_list;
		while (cur) {
			newnode = cur;
			cur = cur->next;
			free(newnode);
		}
	}

	return retval;
}

int main(int argc, char **argv) {
	char log_msg[80];

	int mtype;
	int channel;
	char *message;

	struct mq_attr attrs;

	struct cnx_info_t info;

	struct option long_options[] = 
		{
			{"daemonize", no_argument, 0, 'd'},
			{"mac-addr", required_argument, 0, 'm'},
			{"port", required_argument, 0, 'p'},
			{"wt-path", required_argument, 0, 'w'},
			{"lock-file", required_argument, 0, 'l'},
			{"mq-name", required_argument, 0, 'q'},
			{"mq-timeout", required_argument, 0, 't'},
			{"help", no_argument, 0, 'h'},
			{0, 0, 0, 0}
		};

	int option_index;
	int c;

	memset(lockfile, 0, 128);
	memset(wt_path, 0, 128);
	memset(mq_name, 0, 128);
	memset(ap_mac, 0, 18);
	is_daemon = 1;
	port = -1;
	mq_to = -1;

	while (1) {
		option_index = 0;
		c = getopt_long (argc, argv, "dm:p:w:l:q:t:h", long_options, &option_index);

		if (c == -1) {
			break;
		}

		switch(c) {
			case 'd':
				is_daemon = 1;
				break;
				
			case 'm':
				if (strlen(optarg) != 17) {
					fprintf(stderr, "Invalid mac address for the AP\n\n");
					print_usage();
					exit(1);
				}
				strncpy(ap_mac, optarg, 17);
				break;

			case 'p':
				port = (short)strtol(optarg, NULL, 10);
				if (errno) {
					port = DEF_PORT;
				}
				break;

			case 'w':
				strncpy(wt_path, optarg, 127);
				break;

			case 'l':
				strncpy(lockfile, optarg, 127);
				break;
				
			case 'q':
				strncpy(mq_name, optarg, 127);
				break;

			case 't':
				mq_to = (int)strtol(optarg, NULL, 10);
				if (errno) {
					mq_to = DEF_MQ_TO;
				}
				break;

			case 'h':
				print_usage();
				exit(0);

			case '?':
				fprintf(stderr, "Unknown option\n\n");
				print_usage();
				exit(1);

			default:
				print_usage();
				abort();
		}
	}

	/* argument check */
	if (!*ap_mac) {
		fprintf(stderr, "No AP MAC address given\n\n");
		print_usage();
		exit(1);
	}

	if (!*mq_name) {
		strncpy(mq_name, DEF_DOT11KD_MQNAME, 127);
	}

	if (!*wt_path) {
		strncpy(wt_path, DEF_WT_PATH, 127);
	}

	if (!*lockfile) {
		strncpy(lockfile, DEF_LOCK_FILE, 127);
	}

	if (port == -1) {
		port = DEF_PORT;
	}

	if (mq_to == -1) {
		mq_to = DEF_MQ_TO;
	}

	info.proto = _PROTO_UDP_;
	info.host = "0.0.0.0"; /* any addr */
	info.port = port;

	openlog ("11kd", LOG_PID, LOG_LOCAL5);

	if (is_daemon) {	
		daemonize();
	}

	if (init_server(&info) < 0) {
		log_message(LOG_ERR, "Init failed");
		exit(1);
	}

	/* filling the attribute structure */
	attrs.mq_flags = 0;
	attrs.mq_maxmsg = 100;
	attrs.mq_msgsize = 2048;
	attrs.mq_curmsgs = 123; /* ignored */

	/* open message queue */
	mq = mq_open(mq_name, O_RDWR | O_CREAT, S_IRWXU | S_IRWXG, &attrs);

	if (mq == (mqd_t)-1) {
		fprintf(stderr, "main: Could not create message queue\n");
		fprintf(stderr, "%s\n", strerror(errno));
	}

	while (1) {
		/* read a protocol message */
		message = recv_protocol_message(&info, &mtype, 0, FROM_CLIENT);

		/* check message type and respond */
		if (mtype == MTYPE_REPORT_REQUEST) {
			if (message) free(message);
			log_message(LOG_INFO, "Received report request. Sending back BSS list");

			/* 
			  Handle a request for reports
			 (Cause 11k req to be sent to all stations, receive data from hostap
			 and create an RBAT message)
			*/
			message = handle_collector_request();

			/* send RBAT to collectord */
			xmit_protocol_message(&info, (void *)message, TO_CLIENT);
			if (message) free(message);
		}
		else
		if (mtype == MTYPE_NEW_CHANNEL) {
			/* receive a channel assignment from collectord (TODO) */
			channel = parse_channel_assignment(message);
			memset(log_msg, 0, 80);
			sprintf(log_msg, "Changing channel to %d", channel);
			log_message(LOG_INFO, log_msg);
		}
	}

	return 0;
}

