/**
 * protocol.c -- Implementation of a simple text-based protocol for the 
 * communication between collectord and 11kd. 
 *
 * Copyright (C) 2011 Pantelis A. Frangoudis <pfrag@aueb.gr>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */


#include "protocol.h"
#include "netfunc.h"

/**
 * Converts the message body to lower case
 */
char *msg_to_lower_case(char *msg, int len) {
	char *retval;
	int i;

	retval = (char *)malloc(len + 1);
	memset(retval, 0, len + 1);
	
	memcpy(retval, msg, 4);

	for (i = 4; i < len; i++) {
		retval[i] = (char)tolower((unsigned char)msg[i]);
	}
	return retval;
}

/**
 * Peek connection and retrieve message info. Then, read the appropriate
 * amount of data and message type return the message string for parsing.
 */
char *recv_protocol_message(struct cnx_info_t *info, int *mtype, int timeout, int from_client) {
	int clen;
	int msize;
	char data[PEEK_DATA_LEN];
	char *ldata;
	memset(data, 0, PEEK_DATA_LEN);

	char *message;

	int hdr_len = peek_data(info, data, PEEK_DATA_LEN, timeout, from_client);
	if (hdr_len < 4) {
		return NULL;
	}

	ldata = msg_to_lower_case(data, hdr_len);
	if (!ldata) {
		return NULL;
	}

	/* read message type */
	if (!strncasecmp(data, RBAT_HDR, strlen(RBAT_HDR))) {
		*mtype = MTYPE_REPORT_BATCH;
		if (sscanf(ldata, RBAT_HDR"\r\ncontent-length: %d", &clen) != 1) {
			free(ldata);
			return NULL;
		}
		msize = strlen(RBAT_HDR"\r\ncontent-length: ") + log10(clen) + 3 + clen;
	}
	else if (!strncasecmp(data, CASS_HDR, strlen(CASS_HDR))) {
		*mtype = MTYPE_NEW_CHANNEL;
		if (sscanf(ldata, CASS_HDR"\r\ncontent-length: %d", &clen) != 1) {
			free(ldata);
			return NULL;
		}
		msize = strlen(CASS_HDR"\r\ncontent-length: ") + log10(clen) + 3 + clen;		
	}
	else if (!strncasecmp(data, RREQ_HDR, strlen(RREQ_HDR))) {
		*mtype = MTYPE_REPORT_REQUEST;
		clen = 0;
		msize = 6;
	}
	else {
		free(ldata);
		return NULL;
	}

	if (ldata) free(ldata);

	message = (char*)malloc(msize + 1);
	memset(message, 0, msize + 1);
	
	/* read message */
	if (read_data(info, message, msize, from_client) < 1) {
		if (message) free(message);
		return NULL;
	}
	return message;
}

/**
 * Just call the lower-level write_data method to send a message.
 */
int xmit_protocol_message(struct cnx_info_t *info, char *message, int to_server) {
	return write_data(info, (void *)message, strlen(message), to_server);
}

/**
 * Generate an RREQ message. This is an "empty" message,
 * with no content-length header.
 */
char *generate_report_request() {
	char *retval = (char *)malloc(7);
	memset(retval, 0, 7);
	strncpy(retval, RREQ_HDR"\r\n", 6);

	return retval;
}

/**
 * Parses an RREQ message (dummy func, for now)
 */
int parse_report_request(char *message) {
	if (!message || strlen(message) < 6) {
		return -1;
	}
	if (strncasecmp(RREQ_HDR"\r\n", message, 6)) {
		return 0;
	}
	else {
		return -2;
	}
}

/**
 * Generates a report batch 
 */
char *generate_report_batch(struct bss_info_t* list) {
	int num_bss;
	int mlen;
	int clen;
	int pos;
	char *retval;
	char str_num_bss[5];

	struct bss_info_t* cur = list;
	mlen = 0;
	clen = 0;


	mlen += strlen(RBAT_HDR"\r\nContent-length: "); /* clen will be added at the end */
	mlen += strlen("\r\nNum-bss: ");

	clen += strlen("\r\nNum-bss: "); /* start counting content length */

	num_bss = 0;
	while (cur != NULL) {
		num_bss++;

		mlen += ((cur->channel >= 10) ? 40 : 39);
		clen += ((cur->channel >= 10) ? 40 : 39);

		cur = cur->next;
	}

	mlen += (int)log10(clen) + 1; /* content-length digits */

	memset(str_num_bss, 0, 5);
	sprintf(str_num_bss, "%d", num_bss);
	mlen += strlen(str_num_bss) + 2; /* +2: crlf */

	retval = (char *)malloc(mlen + 1);
	memset(retval, 0, mlen + 1);

	pos = sprintf(retval, RBAT_HDR"\r\nContent-Length: %d\r\nNum-bss: %s\r\n", clen, str_num_bss);

	cur = list;
	while (cur != NULL) {
		pos += sprintf(
			(char*)(retval + pos), 
			"%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x/%2.2x:%2.2x:%2.2x:%2.2x:%2.2x:%2.2x/%d\r\n", 
			cur->user_mac[0], cur->user_mac[1], cur->user_mac[2], cur->user_mac[3], cur->user_mac[4], cur->user_mac[5],
			cur->bss_mac[0], cur->bss_mac[1], cur->bss_mac[2], cur->bss_mac[3], cur->bss_mac[4], cur->bss_mac[5], 
			cur->channel
		);
		cur = cur->next;
	}

	return retval;
}

/**
 * Parses an RBAT message and returns a list of reported BSSs.
 */
struct bss_info_t* parse_report_batch(char *message) {
	int pos;
	int i;
	int num_bss;
	int clen;
	char *lmessage;

	if (!message || strlen(message) < strlen(RBAT_HDR"\r\nContent-Length: ")) {
		return NULL;
	}

	if (strncasecmp(message, RBAT_HDR"\r\nContent-Length: ", strlen(RBAT_HDR"\r\nContent-Length: "))) {
		return NULL;
	}

	/* convert message (except header) to lower case */
	lmessage = msg_to_lower_case(message, strlen(message));
	if (!lmessage) {
		return NULL;
	}

	/* parse content-length and num-bss */
	if (sscanf(lmessage, RBAT_HDR"\r\ncontent-length: %d\r\nnum-bss: %d", &clen, &num_bss) != 2) {
		free(lmessage);
		return NULL;
	}

	pos = strlen(RBAT_HDR"\r\ncontent-length: ") + (int)log10(clen) + 1 + strlen("\r\nnum-bss: ") + (int)log10(num_bss) + 1;

	struct bss_info_t *prev = NULL;
	struct bss_info_t *bss = NULL;
	struct bss_info_t *retval = NULL;

	for (i = 0; i < num_bss; i++) {
		bss = (struct bss_info_t *)malloc(sizeof(struct bss_info_t));
		if (prev) {
			prev->next = bss;
		}
		else {
			retval = bss;
		}

		memset(bss, 0, sizeof(struct bss_info_t));

		sscanf(
			lmessage + pos, 
			"%2hhx:%2hhx:%2hhx:%2hhx:%2hhx:%2hhx/%2hhx:%2hhx:%2hhx:%2hhx:%2hhx:%2hhx/%d\r\n", 
			&(bss->user_mac[0]), &(bss->user_mac[1]), &(bss->user_mac[2]), &(bss->user_mac[3]), &(bss->user_mac[4]), &(bss->user_mac[5]),
			&(bss->bss_mac[0]), &(bss->bss_mac[1]), &(bss->bss_mac[2]), &(bss->bss_mac[3]), &(bss->bss_mac[4]), &(bss->bss_mac[5]), 
			&(bss->channel)
		);

		pos += 38 + (int)log10(bss->channel) + 1;

		prev = bss;
	}

	if (lmessage) {
		free(lmessage);
	}
	return retval;
}

/**
 * Generates a CASS message.
 */
char *generate_channel_assignment(int channel) {
	int clen;
	int mlen;
	char *message;

	clen = strlen("channel: ") + (int)log10(channel) + 1;
	mlen = strlen(CASS_HDR"\r\nContent-Length: \r\n") + clen + (int)log10(clen) + 1;
	message = (char*)malloc(mlen + 1);
	memset(message, 0, mlen + 1);
	sprintf(message, CASS_HDR"\r\nContent-Length: %d\r\nChannel: %d", clen, channel);
	
	return message;
}

/**
 * Parses a CASS message and returns channel.
 */
int parse_channel_assignment(char *message) {
	int clen;
	int channel;
	char *lmessage;

	lmessage = msg_to_lower_case(message, strlen(message));
	if (!lmessage) {
		return -2;
	}

	if (sscanf(lmessage, CASS_HDR"\r\ncontent-length: %d\r\nchannel: %d", &clen, &channel) != 2) {
		free(lmessage);
		return -1;
	}
	else {
		free(lmessage);
		return channel;
	}
}

int generate_random_mac(void *mac) {
	FILE *fp;
	if( (fp = fopen("/dev/urandom", "r")) == NULL) {
	     return -1;
	}

	fread(mac, 1, 6, fp);

	fclose(fp);

	return 0;
}

void bss_list_free(struct bss_info_t *list) {
	struct bss_info_t *cur;
	struct bss_info_t *todel;

	cur = list;
	todel = NULL;
	while (cur) {
		todel = cur;
		cur = cur->next;
		if (todel) free(todel);
	}
}

